# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'gui.ui'
##
## Created by: Qt User Interface Compiler version 5.15.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *

from .icons_rc import *

class Ui_BusViewerWindow(object):
    def setupUi(self, BusViewerWindow):
        if not BusViewerWindow.objectName():
            BusViewerWindow.setObjectName(u"BusViewerWindow")
        BusViewerWindow.resize(898, 571)
        self.actiondraw = QAction(BusViewerWindow)
        self.actiondraw.setObjectName(u"actiondraw")
        icon = QIcon()
        icon.addFile(u":/Icons/icons/automatic_layout.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actiondraw.setIcon(icon)
        self.actionExpand_nodes = QAction(BusViewerWindow)
        self.actionExpand_nodes.setObjectName(u"actionExpand_nodes")
        icon1 = QIcon()
        icon1.addFile(u":/Icons/icons/plus (gray).svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionExpand_nodes.setIcon(icon1)
        self.actionScrink_nodes = QAction(BusViewerWindow)
        self.actionScrink_nodes.setObjectName(u"actionScrink_nodes")
        icon2 = QIcon()
        icon2.addFile(u":/Icons/icons/minus (gray).svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionScrink_nodes.setIcon(icon2)
        self.actionAdjust_to_window_size = QAction(BusViewerWindow)
        self.actionAdjust_to_window_size.setObjectName(u"actionAdjust_to_window_size")
        icon3 = QIcon()
        icon3.addFile(u":/Icons/icons/resize.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.actionAdjust_to_window_size.setIcon(icon3)
        self.centralwidget = QWidget(BusViewerWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.verticalLayout = QVBoxLayout(self.centralwidget)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.toolsFrame = QFrame(self.centralwidget)
        self.toolsFrame.setObjectName(u"toolsFrame")
        self.toolsFrame.setMaximumSize(QSize(16777215, 40))
        self.toolsFrame.setFrameShape(QFrame.NoFrame)
        self.toolsFrame.setFrameShadow(QFrame.Raised)
        self.gridLayout = QGridLayout(self.toolsFrame)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(8, 0, 8, 0)
        self.drawButton = QPushButton(self.toolsFrame)
        self.drawButton.setObjectName(u"drawButton")
        self.drawButton.setIcon(icon)

        self.gridLayout.addWidget(self.drawButton, 0, 3, 1, 1)

        self.label_2 = QLabel(self.toolsFrame)
        self.label_2.setObjectName(u"label_2")

        self.gridLayout.addWidget(self.label_2, 0, 0, 1, 1)

        self.horizontalSpacer = QSpacerItem(621, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.gridLayout.addItem(self.horizontalSpacer, 0, 4, 1, 1)

        self.busNameLabel = QLabel(self.toolsFrame)
        self.busNameLabel.setObjectName(u"busNameLabel")

        self.gridLayout.addWidget(self.busNameLabel, 0, 5, 1, 1)

        self.levelSpinBox = QSpinBox(self.toolsFrame)
        self.levelSpinBox.setObjectName(u"levelSpinBox")
        self.levelSpinBox.setMinimum(1)

        self.gridLayout.addWidget(self.levelSpinBox, 0, 2, 1, 1)


        self.verticalLayout.addWidget(self.toolsFrame)

        self.editorFrame = QFrame(self.centralwidget)
        self.editorFrame.setObjectName(u"editorFrame")
        self.editorFrame.setFrameShape(QFrame.NoFrame)
        self.editorFrame.setFrameShadow(QFrame.Raised)
        self.verticalLayout_2 = QVBoxLayout(self.editorFrame)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.editorLayout = QVBoxLayout()
        self.editorLayout.setObjectName(u"editorLayout")

        self.verticalLayout_2.addLayout(self.editorLayout)


        self.verticalLayout.addWidget(self.editorFrame)

        BusViewerWindow.setCentralWidget(self.centralwidget)
        self.statusbar = QStatusBar(BusViewerWindow)
        self.statusbar.setObjectName(u"statusbar")
        BusViewerWindow.setStatusBar(self.statusbar)
        self.toolBar = QToolBar(BusViewerWindow)
        self.toolBar.setObjectName(u"toolBar")
        self.toolBar.setMovable(False)
        BusViewerWindow.addToolBar(Qt.TopToolBarArea, self.toolBar)

        self.toolBar.addAction(self.actiondraw)
        self.toolBar.addSeparator()
        self.toolBar.addAction(self.actionAdjust_to_window_size)
        self.toolBar.addAction(self.actionScrink_nodes)
        self.toolBar.addAction(self.actionExpand_nodes)

        self.retranslateUi(BusViewerWindow)

        QMetaObject.connectSlotsByName(BusViewerWindow)
    # setupUi

    def retranslateUi(self, BusViewerWindow):
        BusViewerWindow.setWindowTitle(QCoreApplication.translate("BusViewerWindow", u"MainWindow", None))
        self.actiondraw.setText(QCoreApplication.translate("BusViewerWindow", u"draw", None))
        self.actionExpand_nodes.setText(QCoreApplication.translate("BusViewerWindow", u"Expand nodes", None))
        self.actionScrink_nodes.setText(QCoreApplication.translate("BusViewerWindow", u"Scrink nodes", None))
        self.actionAdjust_to_window_size.setText(QCoreApplication.translate("BusViewerWindow", u"Adjust to window size", None))
#if QT_CONFIG(tooltip)
        self.drawButton.setToolTip(QCoreApplication.translate("BusViewerWindow", u"Draw schematic", None))
#endif // QT_CONFIG(tooltip)
        self.drawButton.setText("")
        self.label_2.setText(QCoreApplication.translate("BusViewerWindow", u"Levels to expand", None))
        self.busNameLabel.setText("")
        self.toolBar.setWindowTitle(QCoreApplication.translate("BusViewerWindow", u"toolBar", None))
    # retranslateUi




# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import pandas as pd
import numpy as np
from matplotlib import pyplot as plt

from GridCal.Engine.Devices.bus import Bus
from GridCal.Engine.Devices.enumerations import BranchType, ConverterControlType

from GridCal.Engine.Devices.editable_device import EditableDevice, DeviceType, GCProp


class VSC(EditableDevice):

    def __init__(self, bus_from: Bus = None, bus_to: Bus = None, name='VSC', idtag=None, code='', active=True,
                 r1=0.0001, x1=0.05,
                 m=1.0, m_max=1.1, m_min=0.8,
                 theta=0.1, theta_max=6.28, theta_min=-6.28,
                 Beq=0.001, Beq_min=-0.1, Beq_max=0.1,
                 G0=1e-5, rate=1e-9, kdp=-0.05, k=1.0,
                 control_mode: ConverterControlType = ConverterControlType.type_0_free,
                 Pfset = 0.0, Qfset=0.0, Vac_set=1.0, Vdc_set=1.0,
                 alpha1=0.0001, alpha2=0.015, alpha3=0.2,
                 mttf=0, mttr=0, cost=100, cost_prof=None, rate_prof=None, active_prof=None, contingency_factor=1.0,
                 contingency_enabled=True, monitor_loading=True, contingency_factor_prof=None):
        """
        Voltage source converter (VSC)
        :param bus_from:
        :param bus_to:
        :param name:
        :param idtag:
        :param active:
        :param r1:
        :param x1:
        :param m:
        :param m_max:
        :param m_min:
        :param theta:
        :param theta_max:
        :param theta_min:
        :param G0:
        :param Beq:
        :param Beq_min:
        :param Beq_max:
        :param Inom:
        :param rate:
        :param kdp:
        :param control_mode:
        :param Pfset:
        :param Vac_set:
        :param Vdc_set:
        :param Qfset:
        :param alpha1:
        :param alpha2:
        :param alpha3:
        :param mttf:
        :param mttr:
        :param cost:
        :param cost_prof:
        :param rate_prof:
        :param active_prof:
        """

        EditableDevice.__init__(self,
                                name=name,
                                idtag=idtag,
                                active=active,
                                code=code,
                                device_type=DeviceType.VscDevice,
                                editable_headers={'name': GCProp('', str, 'Name of the branch.'),
                                                  'idtag': GCProp('', str, 'Unique ID', False),
                                                  'code': GCProp('', str, 'Secondary ID'),
                                                  'bus_from': GCProp('', DeviceType.BusDevice,
                                                                     'Name of the bus at the "DC" side of the branch.'),
                                                  'bus_to': GCProp('', DeviceType.BusDevice,
                                                                   'Name of the bus at the "AC" side of the branch.'),
                                                  'active': GCProp('', bool, 'Is the branch active?'),
                                                  'rate': GCProp('MVA', float, 'Thermal rating power of the branch.'),

                                                  'contingency_factor': GCProp('p.u.', float,
                                                                               'Rating multiplier for contingencies.'),
                                                  'contingency_enabled': GCProp('', bool,
                                                                                'Consider this VSC for contingencies.'),
                                                  'monitor_loading': GCProp('', bool,
                                                                            'Monitor this device loading for optimization, NTC or contingency studies.'),
                                                  'mttf': GCProp('h', float, 'Mean time to failure, '
                                                                 'used in reliability studies.'),
                                                  'mttr': GCProp('h', float, 'Mean time to recovery, '
                                                                 'used in reliability studies.'),
                                                  'R1': GCProp('p.u.', float, 'Resistive losses.'),
                                                  'X1': GCProp('p.u.', float, 'Magnetic losses.'),
                                                  'G0': GCProp('p.u.', float, 'Inverter losses.'),

                                                  'Beq': GCProp('p.u.', float, 'Total shunt susceptance.'),
                                                  'Beq_max': GCProp('p.u.', float, 'Max total shunt susceptance.'),
                                                  'Beq_min': GCProp('p.u.', float, 'Min total shunt susceptance.'),

                                                  'm': GCProp('', float, 'Tap changer module, it a value close to 1.0'),
                                                  'm_max': GCProp('', float, 'Max tap changer module'),
                                                  'm_min': GCProp('', float, 'Min tap changer module'),

                                                  'theta': GCProp('rad', float, 'Converter firing angle.'),
                                                  'theta_max': GCProp('rad', float, 'Max converter firing angle.'),
                                                  'theta_min': GCProp('rad', float, 'Min converter firing angle.'),

                                                  'alpha1': GCProp('', float,
                                                                   'Converter losses curve parameter (IEC 62751-2 loss Correction).'),
                                                  'alpha2': GCProp('', float,
                                                                   'Converter losses curve parameter (IEC 62751-2 loss Correction).'),
                                                  'alpha3': GCProp('', float,
                                                                   'Converter losses curve parameter (IEC 62751-2 loss Correction).'),

                                                  'k': GCProp('p.u./p.u.', float, 'Converter factor, typically 0.866.'),

                                                  'control_mode': GCProp('', ConverterControlType,
                                                                         'Converter control mode'),

                                                  'kdp': GCProp('p.u./p.u.', float, 'Droop Power/Voltage slope.'),
                                                  'Pdc_set': GCProp('MW', float, 'DC power set point.'),
                                                  'Qac_set': GCProp('MVAr', float, 'AC Reactive power set point.'),
                                                  'Vac_set': GCProp('p.u.', float, 'AC voltage set point.'),
                                                  'Vdc_set': GCProp('p.u.', float, 'DC voltage set point.'),
                                                  'Cost': GCProp('e/MWh', float, 'Cost of overloads. Used in OPF.'),
                                                  },
                                non_editable_attributes=['bus_from', 'bus_to', 'idtag'],
                                properties_with_profile={'active': 'active_prof',
                                                         'rate': 'rate_prof',
                                                         'contingency_factor': 'contingency_factor_prof',
                                                         'Cost': 'Cost_prof'})

        # the VSC must only connect from an DC to a AC bus
        # this connectivity sense is done to keep track with the articles that set it
        # from -> DC
        # to   -> AC
        # assert(bus_from.is_dc != bus_to.is_dc)
        if bus_to is not None and bus_from is not None:
            # connectivity:
            # for the later primitives to make sense, the "bus from" must be AC and the "bus to" must be DC
            if bus_from.is_dc and not bus_to.is_dc:  # this is the correct sense
                self.bus_from = bus_from
                self.bus_to = bus_to
            elif not bus_from.is_dc and bus_to.is_dc:  # opposite sense, revert
                self.bus_from = bus_to
                self.bus_to = bus_from
                print('Corrected the connection direction of the VSC device:', self.name)
            else:
                raise Exception('Impossible connecting a VSC device here. '
                                'VSC devices must be connected between AC and DC buses')
        else:
            self.bus_from = None
            self.bus_to = None

        # List of measurements
        self.measurements = list()

        # total impedance and admittance in p.u.
        self.R1 = r1
        self.X1 = x1
        self.G0 = G0
        self.Beq = Beq
        self.m = m
        self.theta = theta
        self.k = k
        self.m_max = m_max
        self.m_min = m_min
        self.theta_max = theta_max
        self.theta_min = theta_min
        self.Beq_min = Beq_min
        self.Beq_max = Beq_max

        self.Pdc_set = Pfset
        self.Qac_set = Qfset
        self.Vac_set = Vac_set
        self.Vdc_set = Vdc_set
        self.control_mode = control_mode

        self.kdp = kdp
        self.alpha1 = alpha1
        self.alpha2 = alpha2
        self.alpha3 = alpha3

        self.Cost = cost
        self.Cost_prof = cost_prof

        self.mttf = mttf
        self.mttr = mttr

        self.active = active
        self.active_prof = active_prof

        # branch rating in MVA
        self.rate = rate
        self.contingency_factor = contingency_factor
        self.contingency_enabled: bool = contingency_enabled
        self.monitor_loading: bool = monitor_loading
        self.rate_prof = rate_prof
        self.contingency_factor_prof = contingency_factor_prof

        # branch type: Line, Transformer, etc...
        self.branch_type = BranchType.VSC

    def get_weight(self):
        return np.sqrt(self.R1 * self.R1 + self.X1 * self.X1)

    @property
    def R(self):
        return self.R1

    @property
    def X(self):
        return self.X1

    def change_base(self, Sbase_old, Sbase_new):
        b = Sbase_new / Sbase_old

        self.R1 *= b
        self.X1 *= b
        self.G0 *= b
        self.Beq *= b

    def get_coordinates(self):
        """
        Get the line defining coordinates
        """
        return [self.bus_from.get_coordinates(), self.bus_to.get_coordinates()]

    def correct_buses_connection(self):
        """
        Fix the buses connection (from: DC, To: AC)
        """
        # the VSC must only connect from an DC to a AC bus
        # this connectivity sense is done to keep track with the articles that set it
        # from -> DC
        # to   -> AC
        # assert(bus_from.is_dc != bus_to.is_dc)
        if self.bus_to is not None and self.bus_from is not None:
            # connectivity:
            # for the later primitives to make sense, the "bus from" must be AC and the "bus to" must be DC
            if self.bus_from.is_dc and not self.bus_to.is_dc:  # correct sense
                pass
            elif not self.bus_from.is_dc and self.bus_to.is_dc:  # opposite sense, revert
                self.bus_from, self.bus_to = self.bus_to, self.bus_from
                print('Corrected the connection direction of the VSC device:', self.name)
            else:
                raise Exception('Impossible connecting a VSC device here. '
                                'VSC devices must be connected between AC and DC buses')
        else:
            self.bus_from = None
            self.bus_to = None

    def get_properties_dict(self, version=3):
        """
        Get json dictionary
        :return:
        """

        if self.active_prof is not None:
            active_prof = self.active_prof.tolist()
            rate_prof = self.rate_prof.tolist()
        else:
            active_prof = list()
            rate_prof = list()

        modes = {ConverterControlType.type_0_free: 0,
                 ConverterControlType.type_I_1: 1,
                 ConverterControlType.type_I_2: 2,
                 ConverterControlType.type_I_3: 3,
                 ConverterControlType.type_II_4: 4,
                 ConverterControlType.type_II_5: 5,
                 ConverterControlType.type_III_6: 6,
                 ConverterControlType.type_III_7: 7,
                 ConverterControlType.type_IV_I: 8,
                 ConverterControlType.type_IV_II: 9}
        if version == 2:
            d = {'id': self.idtag,
                 'type': 'vsc',
                 'phases': 'ps',
                 'name': self.name,
                 'name_code': self.code,
                 'bus_from': self.bus_from.idtag,
                 'bus_to': self.bus_to.idtag,
                 'active': self.active,

                 'rate': self.rate,
                 'r': self.R1,
                 'x': self.X1,
                 'g': self.G0,

                 'm': self.m,
                 'm_min': self.m_min,
                 'm_max': self.m_max,

                 'theta': self.theta,
                 'theta_min': self.theta_min,
                 'theta_max': self.theta_max,

                 'Beq': self.Beq,
                 'Beq_min': self.Beq_min,
                 'Beq_max': self.Beq_max,

                 'alpha1': self.alpha1,
                 'alpha2': self.alpha2,
                 'alpha3': self.alpha3,

                 'k': self.k,
                 'kdp': self.kdp,
                 'Pfset': self.Pdc_set,
                 'Qfset': self.Qac_set,
                 'vac_set': self.Vac_set,
                 'vdc_set': self.Vdc_set,

                 'mode': modes[self.control_mode]
                 }
        elif version == 3:
            d = {'id': self.idtag,
                 'type': 'vsc',
                 'phases': 'ps',
                 'name': self.name,
                 'name_code': self.code,
                 'bus_from': self.bus_from.idtag,
                 'bus_to': self.bus_to.idtag,
                 'active': self.active,

                 'rate': self.rate,
                 'contingency_factor1': self.contingency_factor,
                 'contingency_factor2': self.contingency_factor,
                 'contingency_factor3': self.contingency_factor,

                 'r': self.R1,
                 'x': self.X1,
                 'g': self.G0,

                 'm': self.m,
                 'm_min': self.m_min,
                 'm_max': self.m_max,

                 'theta': self.theta,
                 'theta_min': self.theta_min,
                 'theta_max': self.theta_max,

                 'Beq': self.Beq,
                 'Beq_min': self.Beq_min,
                 'Beq_max': self.Beq_max,

                 'alpha1': self.alpha1,
                 'alpha2': self.alpha2,
                 'alpha3': self.alpha3,

                 'k': self.k,
                 'kdp': self.kdp,
                 'Pfset': self.Pdc_set,
                 'Qfset': self.Qac_set,
                 'vac_set': self.Vac_set,
                 'vdc_set': self.Vdc_set,

                 'control_mode': modes[self.control_mode]
                 }
        else:
            d = dict()

        return d

    def get_profiles_dict(self, version=3):
        """

        :return:
        """
        if self.active_prof is not None:
            active_prof = self.active_prof.tolist()
            rate_prof = self.rate_prof.tolist()
        else:
            active_prof = list()
            rate_prof = list()

        return {'id': self.idtag,
                'active': active_prof,
                'rate': rate_prof}

    def get_units_dict(self, version=3):
        """
        Get units of the values
        """
        return {'rate': 'MW',
                'r': 'p.u.',
                'x': 'p.u.',
                'b': 'p.u.',
                'g': 'p.u.'}

    def plot_profiles(self, time_series=None, my_index=0, show_fig=True):
        """
        Plot the time series results of this object
        :param time_series: TimeSeries Instance
        :param my_index: index of this object in the simulation
        :param show_fig: Show the figure?
        """

        if time_series is not None:
            fig = plt.figure(figsize=(12, 8))

            ax_1 = fig.add_subplot(211)
            ax_2 = fig.add_subplot(212, sharex=ax_1)

            x = time_series.results.time

            # loading
            y = time_series.results.loading.real * 100.0
            df = pd.DataFrame(data=y[:, my_index], index=x, columns=[self.name])
            ax_1.set_title('Loading', fontsize=14)
            ax_1.set_ylabel('Loading [%]', fontsize=11)
            df.plot(ax=ax_1)

            # losses
            y = np.abs(time_series.results.losses)
            df = pd.DataFrame(data=y[:, my_index], index=x, columns=[self.name])
            ax_2.set_title('Losses', fontsize=14)
            ax_2.set_ylabel('Losses [MVA]', fontsize=11)
            df.plot(ax=ax_2)

            plt.legend()
            fig.suptitle(self.name, fontsize=20)

        if show_fig:
            plt.show()



# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import numpy as np
import pandas as pd

from matplotlib import pyplot as plt
from GridCal.Engine.Simulations.result_types import ResultTypes


class ResultsTable:
    """
    Class to populate a Qt table view with data from the results
    """
    def __init__(self, data: np.ndarray, columns, index, palette=None, title='', xlabel='', ylabel='', units='',
                 editable=False, editable_min_idx=-1, decimals=6):
        """

        :param data:
        :param columns:
        :param index:
        :param palette:
        :param title:
        :param xlabel:
        :param ylabel:
        :param editable:
        :param editable_min_idx:
        :param decimals:
        """
        if len(data.shape) == 1:
            self.data_c = data.reshape(-1, 1)
        else:
            self.data_c = data
        self.cols_c = columns
        self.index_c = index

        self.editable = editable
        self.editable_min_idx = editable_min_idx
        self.palette = palette
        self.title = title
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.units = units
        self.r, self.c = self.data_c.shape
        self.isDate = False
        if self.r > 0 and self.c > 0:
            if isinstance(self.index_c[0], np.datetime64):
                self.index_c = pd.to_datetime(self.index_c)
                self.isDate = True

        self.format_string = '.' + str(decimals) + 'f'

        self.formatter = lambda x: "%.2f" % x

    def slice_cols(self, col_idx):
        """
        Make column slicing
        :param col_idx: indices of the columns
        :return: Nothing
        """
        sliced_model = ResultsTable(data=self.data_c[:, col_idx],
                                    columns=[self.cols_c[i] for i in col_idx],
                                    index=self.index_c,
                                    palette=None,
                                    title=self.title,
                                    xlabel=self.xlabel,
                                    ylabel=self.ylabel,
                                    units=self.units,
                                    editable=self.editable,
                                    editable_min_idx=self.editable_min_idx,
                                    decimals=6)

        sliced_model.format_string = self.format_string
        return sliced_model

    def slice_rows(self, idx):
        """
        Make rows slicing
        :param idx: indices of the columns
        :return: Nothing
        """
        sliced_model = ResultsTable(data=self.data_c[idx, :],
                                    columns=self.cols_c,
                                    index=[self.index_c[i] for i in idx],
                                    palette=None,
                                    title=self.title,
                                    xlabel=self.xlabel,
                                    ylabel=self.ylabel,
                                    units=self.units,
                                    editable=self.editable,
                                    editable_min_idx=self.editable_min_idx,
                                    decimals=6)

        sliced_model.format_string = self.format_string
        return sliced_model

    def slice_all(self, row_idx, col_idx):
        """
        Make rows slicing
        :param row_idx: indices of the rows
        :param col_idx: indices of the columns
        :return: Nothing
        """
        sliced_model = ResultsTable(data=self.data_c[row_idx, :][:, col_idx],
                                    columns=[self.cols_c[i] for i in col_idx],
                                    index=[self.index_c[i] for i in row_idx],
                                    palette=None,
                                    title=self.title,
                                    xlabel=self.xlabel,
                                    ylabel=self.ylabel,
                                    units=self.units,
                                    editable=self.editable,
                                    editable_min_idx=self.editable_min_idx,
                                    decimals=6)

        sliced_model.format_string = self.format_string
        return sliced_model

    def search_in_columns(self, txt):
        """
        Search stuff
        :param txt:
        :return:
        """
        idx = list()
        txt2 = str(txt).lower()
        for i, val in enumerate(self.cols_c):
            if txt2 in val.lower():
                idx.append(i)
        idx = np.array(idx, dtype=int)
        if len(idx) > 0:
            return self.slice_cols(idx)
        else:
            return None

    def search_in_rows(self, txt):
        """
        Search stuff
        :param txt:
        :return:
        """
        idx = list()
        txt2 = str(txt).lower()
        for i, val in enumerate(self.index_c):
            if txt2 in val.lower():
                idx.append(i)
        idx = np.array(idx, dtype=int)
        if len(idx) > 0:
            return self.slice_rows(idx)
        else:
            return None

    def search(self, txt: str):
        """
        Search stuff
        :param txt:
        :return:
        """
        txt = txt.strip()
        cols = np.array(self.cols_c).astype(str)
        index = np.array(self.index_c).astype(str)

        if txt.startswith('<'):

            txt = txt.replace(' ', '')
            try:
                val = float(txt[1:])
                row_idx, col_idx = np.where(self.data_c < val)

                if len(self.cols_c) == 1:
                    return self.slice_rows(row_idx)
                else:
                    row_idx = np.unique(row_idx)
                    col_idx = np.unique(col_idx)
                    return self.slice_all(row_idx, col_idx)

            except ValueError:
                return None

        elif txt.startswith('>'):

            txt = txt.replace(' ', '')
            try:
                val = float(txt[1:])
                row_idx, col_idx = np.where(self.data_c > val)

                if len(self.cols_c) == 1:
                    return self.slice_rows(row_idx)
                else:
                    row_idx = np.unique(row_idx)
                    col_idx = np.unique(col_idx)
                    return self.slice_all(row_idx, col_idx)

            except ValueError:
                return None

        elif txt.startswith('!='):

            txt = txt.replace(' ', '')
            try:
                val = float(txt[2:])
                row_idx, col_idx = np.where(self.data_c != val)

                if len(self.cols_c) == 1:
                    return self.slice_rows(row_idx)
                else:
                    row_idx = np.unique(row_idx)
                    col_idx = np.unique(col_idx)
                    return self.slice_all(row_idx, col_idx)

            except ValueError:
                return None
        else:
            # search by similar text
            row_idx = list()
            txt2 = str(txt).lower()
            for i, val in enumerate(index):
                if txt2 in val.lower():
                    row_idx.append(i)
            row_idx = np.array(row_idx, dtype=int)

            col_idx = list()
            txt2 = str(txt).lower()
            for i, val in enumerate(cols):
                if txt2 in val.lower():
                    col_idx.append(i)
            col_idx = np.array(col_idx, dtype=int)

            if len(col_idx) > 0:

                if len(row_idx) == 0:
                    # if some col was found but no row, pick all
                    row_idx = np.arange(len(self.index_c))

            else:
                if len(row_idx) > 0:
                    # if some row was found, but no column, pick all
                    col_idx = np.arange(len(self.cols_c))

            return self.slice_all(row_idx, col_idx)

        return None

    def copy_to_column(self, row, col):
        """
        Copies one value to all the column
        @param row: Row of the value
        @param col: Column of the value
        @return: Nothing
        """
        self.data_c[:, col] = self.data_c[row, col]

    def is_complex(self):
        return self.data_c.dtype == complex

    def get_data(self):
        """
        Returns: index, columns, data
        """
        n = len(self.cols_c)

        if n > 0:
            # gather values
            if type(self.cols_c) == pd.Index:
                names = self.cols_c.values

                if len(names) > 0:
                    if type(names[0]) == ResultTypes:
                        names = [str(val) for val in names]
            else:
                names = [str(val) for val in self.cols_c]

            values = self.data_c

            return self.index_c, names, values
        else:
            # there are no elements
            return self.index_c, list(), self.data_c

    def convert_to_cdf(self):
        """
        Convert the data in-place to CDF based
        :return:
        """

        # calculate the proportional values of samples
        n = self.data_c.shape[0]
        if n > 1:
            self.index_c = np.arange(n, dtype=float) / (n - 1)
        else:
            self.index_c = np.arange(n, dtype=float)

        for i in range(self.data_c.shape[1]):
            self.data_c[:, i] = np.sort(self.data_c[:, i], axis=0)

        self.xlabel = 'Probability of value<=x'

    def convert_to_abs(self):
        """
        Convert the data to abs
        :return:
        """
        try:
            self.data_c = np.abs(self.data_c)
        except TypeError:
            print('Could not convert to abs :/')

    def to_df(self):
        """
        get DataFrame
        """
        index, columns, data = self.get_data()

        return pd.DataFrame(data=data, index=index, columns=columns)

    def save_to_excel(self, file_name):
        """
        save data to excel
        :param file_name:
        """
        self.to_df().to_excel(file_name)

    def save_to_csv(self, file_name):
        """
        Save data to csv
        :param file_name:
        """
        self.to_df().to_csv(file_name)

    def get_data_frame(self):
        """
        Save data to csv
        """
        index, columns, data = self.get_data()
        return pd.DataFrame(data=data, index=index, columns=columns)

    def plot(self, ax=None, selected_col_idx=None, selected_rows=None):
        """
        Plot the data model
        :param ax: Matplotlib axis
        :param selected_col_idx: list of selected column indices
        :param selected_rows: list of rows to plot
        """
        index, columns, data = self.get_data()

        if selected_col_idx is not None:
            columns = [columns[i] for i in selected_col_idx]
            data = data[:, selected_col_idx]

        if selected_rows is not None:
            index = [index[i] for i in selected_rows]
            data = data[selected_rows, :]

        if ax is None:
            fig = plt.figure(figsize=(12, 6))
            ax = fig.add_subplot(111)

        if 'voltage' in self.title.lower():
            data[data == 0] = 'nan'  # to avoid plotting the zeros

        if len(columns) > 15:
            plot_legend = False
        else:
            plot_legend = True

        df = pd.DataFrame(data=data, index=index, columns=columns)
        ax.set_title(self.title, fontsize=14)
        ax.set_ylabel(self.ylabel, fontsize=11)
        ax.set_xlabel(self.xlabel, fontsize=11)
        try:
            df.plot(ax=ax, legend=plot_legend)
        except TypeError:
            print('No numeric data to plot...')





# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
import numpy as np
import pandas as pd
# from networkx import DiGraph, all_simple_paths, Graph, all_pairs_dijkstra_path_length
import networkx as nx
from scipy.sparse import lil_matrix, csc_matrix

from typing import List
from sklearn.cluster import DBSCAN
from sklearn.preprocessing import Normalizer

from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.Devices.branch import BranchType
from GridCal.Engine.Devices.bus import Bus
from GridCal.Engine.Simulations.LinearFactors.linear_analysis_driver import LinearAnalysisResults
from GridCal.Engine.Simulations.driver_types import SimulationTypes
from GridCal.Engine.Simulations.driver_template import DriverTemplate

pd.set_option('display.max_rows', 500)
pd.set_option('display.max_columns', 500)
pd.set_option('display.width', 1000)


def get_branches_of_bus(B, j):
    """
    Get the indices of the branches connected to the bus j
    :param B: Branch-bus CSC matrix
    :param j: bus index
    :return: list of branches in the bus
    """
    return [B.indices[k] for k in range(B.indptr[j], B.indptr[j + 1])]


def select_branches_to_reduce(circuit: MultiCircuit, rx_criteria=True, rx_threshold=1e-5,
                              selected_types=BranchType.Branch):
    """
    Find branches to remove
    Args:
        circuit: Circuit to modify in-place
        rx_criteria: use the r+x threshold to select branches?
        rx_threshold: r+x threshold
        selected_types: branch types to select
    """

    branches_to_remove_idx = list()
    branches = circuit.get_branches()
    for i in range(len(branches)):

        # is this branch of the selected type?
        if branches[i].branch_type in selected_types:

            # Am I filtering by r+x threshold?
            if rx_criteria:

                # compute the r+x ratio
                rx = branches[i].R + branches[i].X

                # if the r+x criteria is met, add it
                if rx < rx_threshold:
                    print(i, '->', rx, '<', rx_threshold)
                    branches_to_remove_idx.append(i)

            elif branches[i].branch_type == BranchType.Switch:

                # add switches
                branches_to_remove_idx.append(i)

            else:
                # Add the branch because it was selected and there is no further criteria
                branches_to_remove_idx.append(i)

    return branches_to_remove_idx


def reduce_grid_brute(circuit: MultiCircuit, removed_br_idx):
    """
    Remove the first branch found to be removed.
    this function is meant to be called until it returns false
    Args:
        circuit: Circuit to modify in-place
        removed_br_idx: branch index

    Returns: Nothing
    """

    # form C
    m = circuit.get_branch_number()
    n = len(circuit.buses)
    buses_dict = {bus: i for i, bus in enumerate(circuit.buses)}
    C = lil_matrix((m, n), dtype=int)
    graph = nx.DiGraph()

    # TODO: Fix the topology reduction with the GC example, see what is going on
    branches = circuit.get_branches()
    for i, elm in enumerate(branches):
        # get the from and to bus indices
        f = buses_dict[elm.bus_from]
        t = buses_dict[elm.bus_to]
        graph.add_edge(f, t)
        C[i, f] = 1
        C[i, t] = -1

    C = csc_matrix(C)

    # get branch buses
    bus_f = branches[removed_br_idx].bus_from
    bus_t = branches[removed_br_idx].bus_to
    f = buses_dict[bus_f]
    t = buses_dict[bus_t]

    updated_branches = list()

    # get the number of paths
    n_paths = len(list(nx.all_simple_paths(graph, f, t)))

    if n_paths == 1:  # if there is only one path, merge the buses

        # get the branches that are connected to the bus f
        adjacent_br_idx = get_branches_of_bus(C, f)

        for k, modified_branch in enumerate(adjacent_br_idx):  # for each adjacent branch, reassign the removed bus

            # get the indices of the buses
            f2 = buses_dict[modified_branch.bus_from]
            t2 = buses_dict[modified_branch.bus_to]

            # re-assign the right bus
            if f2 == f:
                modified_branch.bus_from = bus_t
            elif t2 == f:
                modified_branch.bus_to = bus_t

            # copy the state of the removed branch
            modified_branch.active = branches[removed_br_idx].active

            # remember the updated branches
            updated_branches.append(modified_branch)

        # merge buses
        bus_t.merge(bus_f)
        updated_bus = bus_t

        # delete bus
        removed_bus = circuit.buses.pop(f)

        # remove the branch and that's it
        removed_branch = branches.pop(removed_br_idx)

    else:
        # remove the branch and that's it
        removed_branch = branches.pop(removed_br_idx)
        removed_bus = None
        updated_bus = None

    # return the removed branch and the possible removed bus
    return removed_branch, removed_bus, updated_bus, updated_branches


def reduce_buses(circuit: MultiCircuit, buses_to_reduce: List[Bus], text_func=None, prog_func=None):
    """
    Reduce the uses in the grid
    This function removes the buses but whenever a bus is removed, the devices connected to it
    are inherited by the bus of higher voltage that is connected.
    If the bus is isolated, those devices are lost.
    :param circuit: MultiCircuit instance
    :param buses_to_reduce: list of Bus objects
    :return: Nothing
    """

    if text_func is not None:
        text_func('Removing and merging buses...')

    # create dictionary of bus relationships
    bus_bus = dict()
    branches = circuit.get_branches()
    for branch in branches:
        f = branch.bus_from
        t = branch.bus_to

        # add that "t" is related to "f"
        if f in bus_bus.keys():
            bus_bus[f].append(t)
        else:
            bus_bus[f] = [t]

        # add that "f" is related to "t"
        if t in bus_bus.keys():
            bus_bus[t].append(f)
        else:
            bus_bus[t] = [f]

    # sort on voltage
    for bus, related in bus_bus.items():
        related.sort(key=lambda x: x.Vnom, reverse=True)

    buses_merged = list()

    # remove
    total = len(buses_to_reduce)
    for k, bus in enumerate(buses_to_reduce):

        if bus in bus_bus.keys():
            related_buses = bus_bus[bus]

            if len(related_buses) > 0:
                selected = related_buses.pop(0)
                while selected not in circuit.buses and len(related_buses) > 0:
                    selected = related_buses.pop(0)

                # merge the bus with the selected one
                print('Assigning', bus.name, 'to', selected.name)
                selected.merge(bus)

                # merge the graphics
                if selected.graphic_obj is not None and bus.graphic_obj is not None:
                    selected.graphic_obj.merge(bus.graphic_obj)

                # remember the buses that keep the devices
                buses_merged.append(selected)

                # delete the bus from the circuit and the dictionary
                circuit.delete_bus(bus)
                bus_bus.__delitem__(bus)
            else:
                # the bus is isolated, so delete it
                circuit.delete_bus(bus)

        else:
            # the bus is isolated, so delete it
            circuit.delete_bus(bus)

        if text_func is not None:
            text_func('Removing ' + bus.name + '...')

        if prog_func is not None:
            prog_func((k+1) / total * 100.0)

    return buses_merged


class TopologyReductionOptions:

    def __init__(self, rx_criteria=False, rx_threshold=1e-5, selected_types=BranchType.Branch):
        """
        Topology reduction options
        :param rx_criteria:
        :param rx_threshold:
        :param selected_types:
        """

        self.rx_criteria = rx_criteria
        self.rx_threshold = rx_threshold
        self.selected_type = selected_types


class TopologyReduction(DriverTemplate):
    tpe = SimulationTypes.TopologyReduction_run

    def __init__(self, grid: MultiCircuit, branch_indices):
        """
        Topology reduction driver
        :param grid: MultiCircuit instance
        :param options:
        """
        DriverTemplate.__init__(self, grid=grid)

        self.br_to_remove = branch_indices

        self.__cancel__ = False

    def run(self):
        """
        Run the monte carlo simulation
        @return:
        """
        self.progress_signal.emit(0.0)
        self.progress_text.emit('Detecting which branches to remove...')

        # sort the branches in reverse order
        self.br_to_remove.sort(reverse=True)

        total = len(self.br_to_remove)

        # for every branch in reverse order...
        for i, br_idx in enumerate(self.br_to_remove):

            # delete branch
            removed_branch, removed_bus, updated_bus, updated_branches = reduce_grid_brute(circuit=self.grid,
                                                                                           removed_br_idx=br_idx)

            # display progress
            self.progress_text.emit('Removed branch ' + str(br_idx) + ': ' + removed_branch.name)
            progress = (i+1) / total * 100
            self.progress_signal.emit(progress)

    def cancel(self):
        """
        Cancel the simulation
        :return:
        """
        self.__cancel__ = True
        self.progress_signal.emit(0.0)
        self.progress_text.emit('Cancelled')
        self.done_signal.emit()


class DeleteAndReduce(DriverTemplate):

    def __init__(self, grid: MultiCircuit, objects, sel_idx):
        """

        :param grid:
        :param objects: list of objects to reduce (buses in this cases)
        :param sel_idx: indices
        """
        DriverTemplate.__init__(self, grid=grid)

        self.objects = objects

        self.sel_idx = sel_idx

        self.buses_merged = list()

        self.__cancel__ = False

    def run(self):
        """
        Run the monte carlo simulation
        @return:
        """
        self.progress_signal.emit(0.0)
        self.progress_text.emit('Detecting which branches to remove...')

        # get the selected buses
        buses = [self.objects[idx.row()] for idx in self.sel_idx]

        # reduce
        self.buses_merged = reduce_buses(circuit=self.grid,
                                         buses_to_reduce=buses,
                                         text_func=self.progress_text.emit,
                                         prog_func=self.progress_signal.emit)

        # display progress
        self.progress_text.emit('Done')
        self.progress_signal.emit(0.0)
        self.done_signal.emit()

    def cancel(self):
        """
        Cancel the simulation
        :return:
        """
        self.__cancel__ = True
        self.progress_signal.emit(0.0)
        self.progress_text.emit('Cancelled')
        self.done_signal.emit()


class NodeGroupsDriver(DriverTemplate):
    name = 'Node groups'
    tpe = SimulationTypes.NodeGrouping_run

    def __init__(self, grid: MultiCircuit, sigmas=0.5, min_group_size=2, ptdf_results: LinearAnalysisResults = None):
        """
        Electric distance clustering
        :param grid: MultiCircuit instance
        :param sigmas: number of standard deviations to consider
        """
        DriverTemplate.__init__(self, grid=grid)

        self.grid = grid

        self.sigmas = sigmas

        self.min_group_size = min_group_size

        n = len(grid.buses)

        self.use_ptdf = True

        self.ptdf_results = ptdf_results

        # results
        self.X_train = np.zeros((n, n))
        self.sigma = 1.0
        self.groups_by_name = list()
        self.groups_by_index = list()

        self.__cancel__ = False

    def build_weighted_graph(self):
        """

        :return:
        """
        graph = nx.Graph()

        bus_dictionary = {bus: i for i, bus in enumerate(self.grid.buses)}

        for branch_list in self.grid.get_branch_lists():
            for i, branch in enumerate(branch_list):
                # if branch.active:
                f = bus_dictionary[branch.bus_from]
                t = bus_dictionary[branch.bus_to]
                w = branch.get_weight()
                graph.add_edge(f, t, weight=w)

        return graph

    def run(self):
        """
        Run the monte carlo simulation
        @return:
        """
        self.progress_signal.emit(0.0)

        n = len(self.grid.buses)

        if self.use_ptdf:
            self.progress_text.emit('Analyzing PTDF...')

            # the PTDF matrix will be scaled to 0, 1 to be able to train
            self.X_train = Normalizer().fit_transform(self.ptdf_results.PTDF.T)

            metric = 'euclidean'
        else:
            self.progress_text.emit('Exploring Dijkstra distances...')
            # explore
            g = self.build_weighted_graph()
            k = 0
            for i, distances_dict in nx.all_pairs_dijkstra_path_length(g):
                for j, d in distances_dict.items():
                    self.X_train[i, j] = d

                self.progress_signal.emit((k+1) / n * 100.0)
                k += 1
            metric = 'precomputed'

        # compute the sample sigma
        self.sigma = np.std(self.X_train)
        # max_distance = self.sigma * self.sigmas
        max_distance = self.sigmas

        # construct groups
        self.progress_text.emit('Building groups with DBSCAN...')

        # Compute DBSCAN
        model = DBSCAN(eps=max_distance,
                       min_samples=self.min_group_size,
                       metric=metric)

        db = model.fit(self.X_train)

        # get the labels that are greater than -1
        labels = list({i for i in db.labels_ if i > -1})
        self.groups_by_name = [list() for k in labels]
        self.groups_by_index = [list() for k in labels]

        # fill in the groups
        for i, (bus, group_idx) in enumerate(zip(self.grid.buses, db.labels_)):
            if group_idx > -1:
                self.groups_by_name[group_idx].append(bus.name)
                self.groups_by_index[group_idx].append(i)

        # display progress
        self.progress_text.emit('Done')
        self.progress_signal.emit(0.0)
        self.done_signal.emit()

    def cancel(self):
        """
        Cancel the simulation
        :return:
        """
        self.__cancel__ = True
        self.progress_signal.emit(0.0)
        self.progress_text.emit('Cancelled')
        self.done_signal.emit()


if __name__ == '__main__':

    from GridCal.Engine import *
    # fname = '/home/santi/Documentos/GitHub/GridCal/Grids_and_profiles/grids/Illinois 200 Bus.gridcal'
    fname = '/home/santi/Documentos/Private_Grids/Penísula Ibérica 2026.gridcal'
    grid = FileOpen(fname).open()

    driver = NodeGroupsDriver(grid=grid, sigmas=1e-3)
    driver.run()

    print('\nGroups:')
    for group in driver.groups_by_name:
        print(group)

    for group in driver.groups_by_index:
        print(group)



import os
import string
import sys
from random import randint
from enum import Enum
import numpy as np
from numpy.random import default_rng
import networkx as nx
from PySide2.QtWidgets import *

from GridCal.Gui.BusViewer.gui import *
from GridCal.Gui.GridEditorWidget import *
from GridCal.Engine.Devices import *
from GridCal.Engine.Core.multi_circuit import MultiCircuit


class BusViewerGUI(QMainWindow):

    def __init__(self, circuit: MultiCircuit, root_bus: Bus, parent=None, ):
        """

        :param circuit:
        :param root_bus:
        :param parent:
        """
        QMainWindow.__init__(self, parent)
        self.ui = Ui_BusViewerWindow()
        self.ui.setupUi(self)

        self.root_bus = root_bus
        self.bus_id = self.root_bus.name + ' ' + self.root_bus.code

        self.setWindowTitle('Bus Viewer ' + self.bus_id)

        self.circuit = circuit

        # list of bus and branch indices displayed
        self.branch_idx = list()
        self.bus_idx = list()

        # create grid editor o
        self.grid_editor = None

        # create editor and show the root bus
        self.draw()

        # button clicks
        self.ui.drawButton.clicked.connect(self.draw)

        # toolbar clicks
        self.ui.actiondraw.triggered.connect(self.draw)
        self.ui.actionExpand_nodes.triggered.connect(self.bigger_nodes)
        self.ui.actionScrink_nodes.triggered.connect(self.smaller_nodes)
        self.ui.actionAdjust_to_window_size.triggered.connect(self.center_nodes)

    def msg(self, text, title="Warning"):
        """
        Message box
        :param text: Text to display
        :param title: Name of the window
        """
        msg = QMessageBox()
        msg.setIcon(QMessageBox.Information)
        msg.setText(text)
        # msg.setInformativeText("This is additional information")
        msg.setWindowTitle(title)
        # msg.setDetailedText("The details are as follows:")
        msg.setStandardButtons(QMessageBox.Ok)
        retval = msg.exec_()

    def bigger_nodes(self):
        """
        Move the nodes more separated
        """
        if self.grid_editor is not None:
            self.grid_editor.bigger_nodes()

    def smaller_nodes(self):
        """
        Move the nodes closer
        """
        if self.grid_editor is not None:
            self.grid_editor.smaller_nodes()

    def center_nodes(self):
        """
        Center the nodes in the screen
        """
        if self.grid_editor is not None:
            # self.grid_editor.align_schematic()
            self.grid_editor.center_nodes()

    def new_editor(self):
        """
        Create new editor
        """
        self.grid_editor = GridEditor(self.circuit)

        # delete all widgets
        for i in reversed(range(self.ui.editorLayout.count())):
            self.ui.editorLayout.itemAt(i).widget().deleteLater()

        # add the widgets
        self.ui.editorLayout.addWidget(self.grid_editor)

    def draw(self):
        """
        Redraw
        :return:
        """

        self.new_editor()
        self.branch_idx = list()
        self.bus_idx = list()

        bus_dict = self.circuit.get_bus_index_dict()

        # get max expansion level
        max_level = self.ui.levelSpinBox.value()

        # get all branches
        all_branches = self.circuit.get_branches()
        branch_dict = {b: i for i, b in enumerate(all_branches)}

        # sel bus label
        self.ui.busNameLabel.setText(self.bus_id)

        # create a pool of buses
        bus_pool = [(self.root_bus, 0)]  # store the bus objects and their level from the root

        buses = set()
        selected_branches = set()

        while len(bus_pool) > 0:

            # search the next bus
            bus, level = bus_pool.pop()

            self.bus_idx.append(bus_dict[bus])

            # add searched bus
            buses.add(bus)

            if level < max_level:

                for i, br in enumerate(all_branches):

                    if br.bus_from == bus:
                        bus_pool.append((br.bus_to, level + 1))
                        selected_branches.add(br)

                    elif br.bus_to == bus:
                        bus_pool.append((br.bus_from, level + 1))
                        selected_branches.add(br)

                    else:
                        pass

        # sort branches
        lines = list()
        dc_lines = list()
        transformers2w = list()
        hvdc_lines = list()
        vsc_converters = list()
        upfc_devices = list()

        for obj in selected_branches:

            self.branch_idx.append(branch_dict[obj])

            if obj.device_type == DeviceType.LineDevice:
                lines.append(obj)

            elif obj.device_type == DeviceType.DCLineDevice:
                dc_lines.append(obj)

            elif obj.device_type == DeviceType.Transformer2WDevice:
                transformers2w.append(obj)

            elif obj.device_type == DeviceType.HVDCLineDevice:
                hvdc_lines.append(obj)

            elif obj.device_type == DeviceType.VscDevice:
                vsc_converters.append(obj)

            elif obj.device_type == DeviceType.UpfcDevice:
                upfc_devices.append(obj)

            elif obj.device_type == DeviceType.BranchDevice:
                # we need to convert it :D
                obj2 = convert_branch(obj)
                lines.append(obj2)  # call this again, but this time it is not a Branch object
            else:
                raise Exception('Unrecognized branch type ' + obj.device_type.value)

        # Draw schematic subset
        self.grid_editor.add_elements_to_schematic(buses=list(buses),
                                                   lines=lines,
                                                   dc_lines=dc_lines,
                                                   transformers2w=transformers2w,
                                                   hvdc_lines=hvdc_lines,
                                                   vsc_devices=vsc_converters,
                                                   upfc_devices=upfc_devices,
                                                   explode_factor=1.0,
                                                   prog_func=None,
                                                   text_func=print)

        for bus in buses:
            bus.graphic_obj.arrange_children()

        self.center_nodes()


if __name__ == "__main__":

    app = QApplication(sys.argv)
    circuit_ = MultiCircuit()
    circuit_.add_bus(Bus('bus1'))
    window = BusViewerGUI(circuit=circuit_, root_bus=circuit_.buses[0])
    window.resize(1.61 * 700.0, 600.0)  # golden ratio
    window.show()
    sys.exit(app.exec_())




from PySide2.QtGui import QPalette, QColor





def css_rgb(color, a=False):
    """Get a CSS `rgb` or `rgba` string from a `QtGui.QColor`."""
    return ("rgba({}, {}, {}, {})" if a else "rgb({}, {}, {})").format(*color.getRgb())


class QDarkPalette(QPalette):
    """Dark palette for a Qt application meant to be used with the Fusion theme."""
    def __init__(self, *__args):
        super().__init__(*__args)

        self.WHITE = QColor(255, 255, 255)
        self.BLACK = QColor(0, 0, 0)
        self.RED = QColor(255, 0, 0)
        self.PRIMARY = QColor(53, 53, 53)
        self.SECONDARY = QColor(35, 35, 35)
        self.TERTIARY = QColor(42, 130, 218)

        # Set all the colors based on the constants in globals
        self.setColor(QPalette.Window,          self.PRIMARY)
        self.setColor(QPalette.WindowText,      self.WHITE)
        self.setColor(QPalette.Base,            self.SECONDARY)
        self.setColor(QPalette.AlternateBase,   self.PRIMARY)
        self.setColor(QPalette.ToolTipBase,     self.WHITE)
        self.setColor(QPalette.ToolTipText,     self.WHITE)
        self.setColor(QPalette.Text,            self.WHITE)
        self.setColor(QPalette.Button,          self.PRIMARY)
        self.setColor(QPalette.ButtonText,      self.WHITE)
        self.setColor(QPalette.BrightText,      self.RED)
        self.setColor(QPalette.Link,            self.TERTIARY)
        self.setColor(QPalette.Highlight,       self.TERTIARY)
        self.setColor(QPalette.HighlightedText, self.BLACK)

    def set_stylesheet(self, app):
        """
        Static method to set the tooltip stylesheet to a `QtWidgets.QApplication`
        """
        app.setStyleSheet("QToolTip {{"
                          "color: {white};"
                          "background-color: {tertiary};"
                          "border: 1px solid {white};"
                          "}}".format(white=css_rgb(self.WHITE), tertiary=css_rgb(self.TERTIARY)))

    def set_app(self, app):
        """
        Set the Fusion theme and this palette to a `QtWidgets.QApplication`.
        """
        app.setStyle("Fusion")
        app.setPalette(self)
        self.set_stylesheet(app)
